#!/usr/local/bin/tclsh
#
# Convert Ousterhout format man pages into highly crosslinked
# hypertext.
#
# Along the way detect many unmatched font changes and other odd
# things.
#
# Note well, this program is a real hack.  It is only offered as an
# example of how someone might convert a specific set of man pages
# into hypertext, not as a general solution to the problem.  If you
# try to use this, you'll be very much on your own.
#
# Copyright (c) 1995 Roger E. Critchlow Jr
#
# All rights reserved, fair use permitted,
# caveat emptor, and no warranties.
#
# Revisions:
#  May 15, 1995 - initial release
#  May 16, 1995 - added a back to home link to toplevel table of
#	contents.
#  May 18, 1995 - broke toplevel table of contents into separate
#	pages for each section, and broke long table of contents
#	into a one page for each man page.

#
# These variables determine where the man pages come from
# and where the converted pages go to.
#
set tcltkdir /packages
set tkdir tk4.0
set tcldir tcl7.4
set webdir /packages/www.elf.org/tcltk-man-html

##
##
##
set manual(report-level) 1;

proc manerror {msg} {
    global manual;
    set name {};
    set subj {};
    if {[info exists manual(name)]} {
	set name $manual(name);
    }
    if {[info exists manual(section)] && "$manual(section)" != {}} {
	puts stderr "$name: $manual(section):  $msg";
    } else {
	puts stderr "$name: $msg";
    }
}

proc manreport {level msg} {
    global manual;
    if {$level < $manual(report-level)} {
	manerror $msg;
    }
}

proc fatal {msg} {
    global manual;
    manerror $msg;
    exit 1;
}
##
## parsing
##
proc unquote arg {
    regsub -all \" $arg {} arg;
    return $arg;
}

proc parse-directive {line codename restname} {
    upvar $codename code $restname rest;
    return [regexp {^(\.[.a-zA-Z0-9]+) *(.*)} $line all code rest];
}

proc process-text {text} {
    global manual;
    # preprocess text
    regsub -all {\\&} $text \t text;		# some kind of tab?
    regsub -all {&} $text {\&amp;} text;
    regsub -all {\\ } $text {\&nbsp;} text;	# non breaking space
    regsub -all {\\%} $text {} text;		# don't break word following?
    regsub -all "\\\\\n" $text "\n" text;	# 
    regsub -all \" $text {\&quot;} text;
    regsub -all {<} $text {\&lt;} text;
    regsub -all {>} $text {\&gt;} text;
    regsub -all {\\-\\\|\\-} $text -- text;	# two hyphens
    regsub -all -- {\\-\\\^\\-} $text -- text;	# two hyphens
    regsub -all {\\-} $text - text;		# a hyphen
    regsub -all {\\0} $text { } text;		# a space
    # regsub -all {\\\|} $text | text;		# a very thin space 
    regsub -all {\\e} $text {\&#92;} text;	# reverse solidus, ie backslash
    regsub -all {\\\(\+-} $text {\&#177;} text;	# plus or minus sign
    regsub -all {\\fP} $text {\\fR} text;	# a funky font in expr.n
    while {[regexp \\\\ $text]} {
	# C R
	if {[regexp {^([^\\]*)\\fC([^\\]*)\\fR(.*)$} $text all head word tail]} {
	    set text "$head<TT>$word</TT>$tail";
	    continue;
	}
	# B R
	if {[regexp {^([^\\]*)\\fB([^\\]*)\\fR(.*)$} $text all head word tail]} {
	    set text "$head<B>$word</B>$tail";
	    continue;
	}
	# B I
	if {[regexp {^([^\\]*)\\fB([^\\]*)\\fI(.*)$} $text all head word tail]} {
	    set text "$head<B>$word</B>\\fI$tail";
	    continue;
	}
	# B B
	if {[regexp {^([^\\]*)\\fB([^\\]*)\\fB(.*)$} $text all head word tail]} {
	    set text "$head\\fB$word$tail";
	    continue;
	}
	# I R
	if {[regexp {^([^\\]*)\\fI([^\\]*)\\fR(.*)$} $text all head word tail]} {
	    set text "$head<I>$word</I>$tail";
	    continue;
	}
	# I B
	if {[regexp {^([^\\]*)\\fI([^\\]*)\\fB(.*)$} $text all head word tail]} {
	    set text "$head<I>$word</I>\\fB$tail";
	    continue;
	}
	# I I
	if {[regexp {^([^\\]*)\\fI([^\\]*)\\fI(.*)$} $text all head word tail]} {
	    set text "$head\\fI$word$tail";
	    continue;
	}
	manerror "process-text: $text"
	break;
    }
    return $text
}    
##
## pass 2 text input and matching
##
proc open-text {} {
    global manual;
    set manual(text-length) [llength $manual(text)];
    set manual(text-pointer) 0;
}
proc more-text {} {
    global manual;
    return [expr $manual(text-pointer) < $manual(text-length)];
}
proc next-text {} {
    global manual;
    if {[more-text]} {
	set text [lindex $manual(text) $manual(text-pointer)];
	incr manual(text-pointer);
	return $text;
    }
    manerror "read past end of text";
    error "fatal";
}
proc is-a-directive {line} {
    return [expr {[string first . $line] == 0}]
}
proc next-text-is {op restname} {
    global manual;
    upvar $restname rest;
    if {[more-text]} {
	set text [lindex $manual(text) $manual(text-pointer)];
	set top [lindex $text 0];
	if {"$op" == "$top"} {
	    set rest [lrange $text 1 end];
	    incr manual(text-pointer);
	    return 1;
	}
    }
    return 0;
}
proc backup-text {n} {
    global manual;
    if {$manual(text-pointer)-$n >= 0} {
	incr manual(text-pointer) -$n;
    }
}
proc match-text args {
    global manual;
    set nargs [llength $args];
    if {$manual(text-pointer) + $nargs > $manual(text-length)} {
	return 0;
    }
    set nback 0;
    foreach arg $args {
	if { ! [more-text]} {
	    backup-text $nback;
	    return 0;
	}
	set arg [string trim $arg];
	set targ [string trim [lindex $manual(text) $manual(text-pointer)]];
	if {"$arg" == "$targ"} {
	    incr nback;
	    incr manual(text-pointer);
	    continue;
	}
	if {[regexp {^@([_a-zA-Z0-9]+)$} $arg all name]} {
	    upvar $name var;
	    set var $targ;
	    incr nback;
	    incr manual(text-pointer);
	    continue;
	}
	backup-text $nback;
	return 0;
    }
    return 1;
}
proc expand-next-text {n} {
    global manual;
    return [join [lrange $manual(text) $manual(text-pointer) [expr $manual(text-pointer)+$n-1]] \n\n];
}
##
## pass 2 output
##
proc man-puts {text} {
    global manual;
    lappend manual(output-$manual(wing-file)-$manual(name)) $text;
}

##
## build hypertext links to tables of contents
##
proc long-toc {text} {
    global manual;
    set here M[incr manual(section-toc-n)]
    set there L[incr manual(long-toc-n)];
    lappend manual(section-toc) "<LI><A HREF=\"$manual(tail).html#$here\" NAME=\"$there\">$text</A>";
#    return "<A HREF=\"$manual(tail).toc.html#$there\" NAME=\"$here\">$text</A>";
    return "<A NAME=\"$here\">$text</A>";
}
proc option-toc {name class switch} {
    global manual;
    if {"$manual(section)" == {WIDGET-SPECIFIC OPTIONS}} {
	# link the defined option into the long table of contents
	set link [long-toc "$name, $class, $switch"];
	regsub ", $class, $switch" $link {} link;
	return $link;
    } elseif {"$manual(name):$manual(section)" == {options:DESCRIPTION}} {
	# link the defined standard option to the long table of
	# contents and make a target for the standard option references
	# from other man pages.
	set here M$name;
	set there L[incr manual(long-toc-n)];
	set manual(standard-option-$name) "<A HREF=\"$manual(tail).html#$here\">$name, $class, $switch</A>";
	lappend manual(section-toc) "<LI><A HREF=\"$manual(tail).html#$here\" NAME=\"$there\">$name, $class, $switch</A>";
#	return "<A HREF=\"$manual(tail).toc.html#$there\" NAME=\"$here\">$name</A>";
	return "<A NAME=\"$here\">$name</A>";
    } else {
	error "option-toc in $manual(name) section $manual(section)";
    }
}
proc std-option-toc {name} {
    global manual;
    if {[info exists manual(standard-option-$name)]} {
	lappend manual(section-toc) <LI>$manual(standard-option-$name);
	return $manual(standard-option-$name);
    }
    set here M[incr manual(section-toc-n)]
    set there L[incr manual(long-toc-n)];
    set other M$name;
    lappend manual(section-toc) "<LI><A HREF=\"options.n.html#$other\">$name</A>";
    return "<A HREF=\"options.n.html#$other\">$name</A>";
}
##
## process the widget option section
## in widget and options man pages
##
proc output-widget-options {} {
    global manual
    if { ! [match-text {.ta 4c}]} {
	fatal "widget-specific options botch 1: [next-text]";
    }
    man-puts <DL>;
    lappend manual(section-toc) <UL>;
    while {[match-text .PP .nf @head .fi .IP @body]} {
	set head [split $head \n];
	set name [lindex $head 0];
	set class [lindex $head 1];
	set switch [lindex $head 2];
	if { ! [regexp {^Name:	(<.>)([a-zA-Z0-9]+)(</.>)$} $name all oname name cname]} {
	    error "not Name: $name";
	}
	if { ! [regexp {^Class:	(<.>)([a-zA-Z0-9]+)(</.>)$} $class all oclass class cclass]} {
	    error "not Class: $class";
	}
	if { ! [regexp {^Command-Line Switch:	(<.>)([-a-zA-Z0-9 ]+)(</.>)$} $switch all oswitch switch cswitch]} {
	    error "not Switch: $switch";
	}
	man-puts "<DT>Name:	$oname[option-toc $name $class $switch]$cname";
	man-puts "<DT>Class: $oclass$class$cclass";
	man-puts "<DT>Command-Line Switch:	$oswitch$switch$cswitch";
	man-puts <DD>$body;
	if {[next-text-is .PP rest]} {
	    backup-text 1;
	    continue;
	}
	if {[next-text-is .BE rest] || [next-text-is .SH rest]} {
	    backup-text 1;
	    break;
	}
	fatal "widget-specific option botch in loop: [next-text]";
	break;
    }
    man-puts </DL>;
    lappend manual(section-toc) </UL>;
}
##
## process .RS lists
##
proc output-RS-list {} {
    global manual;
    if {[next-text-is .IP rest]} {
	output-IP-list .RS .IP $rest;
	if {[next-text-is .RE rest]} {
	    return;
	}
    }
    man-puts <UL>;
    while {[more-text]} {
	set line [next-text];
	if {[is-a-directive $line]} {
	    set code [lindex $line 0];
	    set rest [lrange $line 1 end];
	    switch -exact $code {
		.RE {
		    break;
		}
		.SH {
		    manerror "unbalanced .RS at section end";
		    backup-text 1;
		    break;
		}
		default {
		    output-directive $line;
		}
	    }
	} else {
	    man-puts $line;
	}
    }	
    man-puts </UL>;
}    
##
## process .IP lists which may be plain indents,
## numeric lists, or definition lists
##
proc output-IP-list {context code rest} {
    global manual;
    if {"$rest" == {}} {
	# blank label, plain indent, no contents entry
	man-puts <UL>
	man-puts "<P>";
	while {[more-text]} {
	    set line [next-text];
	    if {[is-a-directive $line]} {
		set code [lindex $line 0];
		set rest [lrange $line 1 end];
		if {"$code" == {.IP} && "$rest" == {}} {
		    man-puts "<P>";
		    continue;
		}
		if {[lsearch {.br .DS .RS} $code] >= 0} {
		    output-directive $line;
		} else {
		    backup-text 1;
		    break;
		}
	    } else {
		man-puts $line;
	    }
	}
	man-puts </UL>;
    } elseif {"$rest" == {[1]}} {
	# numeric label, ordered list, no contents entry
	man-puts <OL>
	set n 1;
	man-puts "<LI>";
	while {[more-text]} {
	    set line [next-text];
	    if {[is-a-directive $line]} {
		set code [lindex $line 0];
		set rest [lrange $line 1 end];
		if {"$code" == {.IP} && "$rest" == "\[[incr n]]"} {
		    man-puts "<LI>";
		} elseif {"$code" == {.RS}} {
		    output-directive $line;
		    if {"$manual(name)" == {Tcl} && [match-text .PP @stuff]} {
			man-puts <P>;
			man-puts $stuff
		    }
		} elseif {[lsearch {.br .DS} $code] >= 0} {
		    output-directive $line;
		} else {
		    backup-text 1;
		    break;
		}
	    } else {
		man-puts $line;
	    }
	}
	man-puts </OL>;
    } else {
	# labelled list, make contents
	if {"$context" != {.SH}} {
	    man-puts <P>;
	}
	man-puts <DL>
	lappend manual(section-toc) <UL>;
	if {"$manual(section)" == {ARGUMENTS}} {
	    man-puts "<DT>$rest<DD>";
	} else {
	    man-puts "<DT>[long-toc $rest]<DD>";
	}
	if {"$manual(name):$manual(section)" == {selection:DESCRIPTION}} {
	    if {[match-text .RE @rest .RS .RS]} {
		man-puts <DT>[long-toc $rest]<DD>;
	    }
	}
	while {[more-text]} {
	    set line [next-text];
	    if {[is-a-directive $line]} {
		set code [lindex $line 0];
		set rest [lrange $line 1 end];
		if {"$code" == {.IP}} {
		    if {"$manual(section)" == {ARGUMENTS}} {
			man-puts "<DT>$rest<DD>";
		    } else {
			man-puts "<DT>[long-toc $rest]<DD>";
		    }
		    if {"$manual(name):$manual(section)" == {selection:DESCRIPTION}} {
			if {[match-text .RE @rest .RS .RS]} {
			    man-puts <DT>[long-toc $rest]<DD>;
			}
		    }
		} elseif {[lsearch {.br .DS .RS} $code] >= 0} {
		    output-directive $line;
		} else {
		    backup-text 1;
		    break;
		}
	    } else {
		man-puts $line;
	    }
	}
	man-puts </DL>;
	lappend manual(section-toc) </UL>;
    }
}
##
## handle the NAME section lines
## there's only one line in the NAME section,
## consisting of a comma separated list of names,
## followed by a hyphen and a short description.
##
proc output-name {line} {
    global manual;
    # split name line into pieces
    regexp {^([^-]+) - (.*)$} $line all head tail;
    # output line to manual page untouched
    man-puts $line
    # output line to long table of contents
    lappend manual(long-toc) <DL>;
    foreach name [split $head ,] {
	set name [string trim $name];
	lappend manual(long-toc) "<A HREF=\"$manual(tail).html\" NAME=\"$manual(name)\"><DT>$name</A>"
    }
    lappend manual(long-toc) <DD>$tail</DL>;
    # separate out the names for future reference
    foreach name [split $head ,] {
	set name [string trim $name];
	if {[llength $name] > 1} {
	    manerror "name has a space: {$name}\nfrom: $line";
	}
	lappend manual(wing-toc) $name;
	lappend manual(name-$name) $manual(wing-file)/$manual(tail);
    }
}
##
## build a cross-reference link if appropriate
##
proc cross-reference {ref} {
    global manual;
    if {[string match Tcl_* $ref]} {
	set lref $ref;
    } elseif {[string match Tk_* $ref]} {
	set lref $ref;
    } elseif {"$ref" == {Tcl}} {
	set lref $ref;
    } else {
	set lref [string tolower $ref];
    }
    ##
    ## nothing to reference
    ##
    if { ! [info exists manual(name-$lref)]} {
	if {"$lref" == {string match} && $manual(tail) != {string.n}} {
	    return "<A HREF=\"../$manual(name-string).html\">$ref</A>";
	}
	if {"$lref" == {image create} && $manual(tail) != {image.n}} {
	    return "<A HREF=\"../$manual(name-image).html\">$ref</A>";
	}
	if {[lsearch {stdin stdout stderr end} $lref] >= 0} {
	    # no good place to send these
	    # tcl tokens?
	    # also end
	}
	return $ref;
    }
    ##
    ## multiple choices for reference
    ##
    if {[llength $manual(name-$lref)] > 1} {
	if {$lref == {exit}} {
	    if {"$manual(tail)" == {tclsh.1}} {
		set i [lsearch -glob $manual(name-$lref) *TclCmd*];
		if {$i >= 0} {
		    set exit [lindex $manual(name-$lref) $i];
		    return "<A HREF=\"../$exit.html\">$ref</A>"
		}
	    }
	    if {$manual(tail) == {exit.n} && $manual(wing-file) == {TkCmd}} {
		set i [lsearch -glob $manual(name-$lref) *TclCmd*];
		if {$i >= 0} {
		    set exit [lindex $manual(name-$lref) $i];
		    return "<A HREF=\"../$exit.html\">$ref</A>"
		}
	    }
	}
	puts stderr "multiple cross reference to $ref in $manual(name-$lref) from $manual(wing-file)/$manual(tail)";
	return $ref;
    }
    ##
    ## would be a self reference
    ##
    if {[lsearch $manual(name-$lref) $manual(wing-file)/$manual(tail)] >= 0} {
	return $ref;
    }
    ##
    ## exceptions, sigh, to the rule
    ##
    switch $manual(tail) {
	canvas.n {
	    if {$lref == {focus}} {
		upvar tail tail;
		set clue [string first command $tail];
		if {$clue < 0 ||  $clue > 5} {
		    return $ref;
		}
	    }
	    if {[lsearch {bitmap image text} $lref] >= 0} {
		return $ref;
	    }
	}
	checkbutton.n -
	radiobutton.n {
	    if {[lsearch {image} $lref] >= 0} {
		return $ref;
	    }
	}
	menu.n {
	    if {[lsearch {checkbutton radiobutton} $lref] >= 0} {
		return $ref;
	    }
	}
	options.n {
	    if {[lsearch {bitmap image set} $lref] >= 0} {
		return $ref;
	    }
	}
	regexp.n {
	    if {[lsearch {string} $lref] >= 0} {
		return $ref;
	    }
	}
	history.n {
	    if {[lsearch {exec} $lref] >= 0} {
		return $ref;
	    }
	}
	return.n {
	    if {[lsearch {error continue break} $lref] >= 0} {
		return $ref;
	    }
	}
	scrollbar.n {
	    if {[lsearch {set} $lref] >= 0} {
		return $ref;
	    }
	}
    }
    ##
    ## return the cross reference
    ##
    return "<A HREF=\"../$manual(name-$lref).html\">$ref</A>";
}
##
## reference generation errors
##
proc reference-error {msg text} {
    global manual;
    puts stderr "$manual(tail): $msg: {$text}";
    return $text;
}
##
## insert as many cross references into this text string as are appropriate
##
proc insert-cross-references {text} {
    global manual;
    ##
    ## we identify cross references by:
    ##     ``quotation''
    ##    <B>emboldening</B>
    ##    Tcl_ prefix
    ##    Tk_ prefix
    ##	  [a-zA-Z0-9]+ manual entry
    ## and we avoid messing with already anchored text
    ##
    ##
    ## find where each item lives
    ##
    array set offset [list \
			  anchor [string first {<A } $text] \
			  end-anchor [string first {</A>} $text] \
			  quote [string first {``} $text] \
			  end-quote [string first {''} $text] \
			  bold [string first {<B>} $text] \
			  end-bold [string first {</B>} $text] \
			  tcl [string first {Tcl_} $text] \
			  tk [string first {Tk_} $text] \
			  Tcl1 [string first {Tcl manual entry} $text] \
			  Tcl2 [string first {Tcl overview manual entry} $text] \
		     ];
    ##
    ## accumulate a list
    ##
    foreach name [array names offset] {
	if {$offset($name) >= 0} {
	    set invert($offset($name)) $name;
	    lappend offsets $offset($name);
	}
    }
    ##
    ## if nothing, then we're done.
    ##
    if { ! [info exists offsets]} {
	return $text;
    }
    ##
    ## sort the offsets
    ##
    set offsets [lsort -integer $offsets];
    ##
    ## see which we want to use
    ##
    switch -exact $invert([lindex $offsets 0]) {
	anchor {
	    if {$offset(end-anchor) < 0} { return [reference-error {Missing end anchor} $text]; }
	    set head [string range $text 0 $offset(end-anchor)];
	    set tail [string range $text [expr $offset(end-anchor)+1] end];
	    return $head[insert-cross-references $tail]
	}
	quote {
	    if {$offset(end-quote) < 0} { return [reference-error {Missing end quote} $text]; }
	    if {"$invert([lindex $offsets 1])" == {tk}} { set offsets [lreplace $offsets 1 1]; }
	    if {"$invert([lindex $offsets 1])" == {tcl}} { set offsets [lreplace $offsets 1 1]; }
	    switch -exact $invert([lindex $offsets 1]) {
		end-quote {
		    set head [string range $text 0 [expr $offset(quote)-1]]
		    set body [string range $text [expr $offset(quote)+2] [expr $offset(end-quote)-1]];
		    set tail [string range $text [expr $offset(end-quote)+2] end];
		    return $head``[cross-reference $body]''[insert-cross-references $tail];
		}
		bold -
		anchor {
		    set head [string range $text 0 [expr $offset(end-quote)+1]]
		    set tail [string range $text [expr $offset(end-quote)+2] end];
		    return $head[insert-cross-references $tail];
		}
	    }
	    return [reference-error {Uncaught quote case} $text];
	}
	bold {
	    if {$offset(end-bold) < 0} { return $text; }
	    if {"$invert([lindex $offsets 1])" == {tk}} { set offsets [lreplace $offsets 1 1]; }
	    if {"$invert([lindex $offsets 1])" == {tcl}} { set offsets [lreplace $offsets 1 1]; }
	    switch -exact $invert([lindex $offsets 1]) {
		end-bold {
		    set head [string range $text 0 [expr $offset(bold)-1]]
		    set body [string range $text [expr $offset(bold)+3] [expr $offset(end-bold)-1]];
		    set tail [string range $text [expr $offset(end-bold)+4] end];
		    return $head<B>[cross-reference $body]</B>[insert-cross-references $tail];
		}
		anchor {
		    set head [string range $text 0 [expr $offset(end-bold)+3]]
		    set tail [string range $text [expr $offset(end-bold)+4] end];
		    return $head[insert-cross-references $tail];
		}
	    }
	    return [reference-error {Uncaught bold case} $text];
	}
	tk {
	    set head [string range $text 0 [expr $offset(tk)-1]]
	    set tail [string range $text $offset(tk) end];
	    if { ! [regexp {^(Tk_[a-zA-Z0-9_]+)(.*)$} $tail all body tail]} { return [reference-error {Tk regexp failed} $text]; }
	    return $head[cross-reference $body][insert-cross-references $tail];
	}
	tcl {
	    set head [string range $text 0 [expr $offset(tcl)-1]]
	    set tail [string range $text $offset(tcl) end];
	    if { ! [regexp {^(Tcl_[a-zA-Z0-9_]+)(.*)$} $tail all body tail]} { return [reference-error {Tcl regexp failed} $text]; }
	    return $head[cross-reference $body][insert-cross-references $tail];
	}
	Tcl1 -
	Tcl2 {
	    set off [lindex $offsets 0];
	    set head [string range $text 0 [expr $off-1]];
	    set body Tcl
	    set tail [string range $text [expr $off+3] end];
	    return $head[cross-reference $body][insert-cross-references $tail];
	}
	end-anchor -
	end-bold -
	end-quote {
	    return [reference-error "Out of place $invert([lindex $offsets 0])" $text];
	}
    }
}
##
## process formatting directives
##
proc output-directive {line} {
    global manual;
    # process format directive
    set code [lindex $line 0];
    set rest [lrange $line 1 end];
    switch -exact $code {
	.BS -
	.BE {
	    # man-puts <HR>;
	}
	.SH {
	    # drain any open lists
	    # announce the subject
	    set manual(section) $rest;
	    # start our own stack of stuff
	    set manual($manual(name)-$manual(section)) {};
	    lappend manual(has-$manual(section)) $manual(name);
	    man-puts "<H3>[long-toc $manual(section)]</H3>";
	    # some sections can simply free wheel their way through the text
	    # some sections can be processed in their own loops
	    switch -exact $manual(section) {
		NAME {
		    while {1} {
			set line [next-text];
			if {[is-a-directive $line]} {
			    backup-text 1;
			    return;
			}
			output-name $line;
		    }
		}
		SYNOPSIS {
		    lappend manual(section-toc) <UL>;
		    while {1} {
			if {[next-text-is .nf rest]
			 || [next-text-is .br rest]
			 || [next-text-is .fi rest]} {
			    continue;
			}
			if {[next-text-is .SH rest]
		         || [next-text-is .BE rest]} {
			    backup-text 1;
			    break;
			}
			if {[next-text-is .sp rest]} {
			    man-puts <P>;
			    continue;
			}
			set more [next-text];
			if {[is-a-directive $more]} {
			    manerror "in SYNOPSIS found $more";
			    backup-text 1;
			    break;
			} else {
			    foreach more [split $more \n] {
				man-puts $more<BR>;
				lappend manual(section-toc) $more<BR>;
			    }
			}
		    }
		    lappend manual(section-toc) </UL>;
		    return;
		}
		{SEE ALSO} {
		    while {[more-text]} {
			if {[next-text-is .SH rest]} {
			    backup-text 1;
			    return;
			}
			set more [next-text];
			if {[is-a-directive $more]} {
			    manerror "$more";
			    backup-text 1;
			    return;
			}
			set nmore {};
			foreach cr [split $more ,] {
			    set cr [string trim $cr];
			    if { ! [regexp {^<B>.*</B>$} $cr]} {
				set cr <B>$cr</B>;
			    }
			    lappend nmore $cr;
			}
			man-puts [join $nmore {, }];
		    }
		    return;
		}
		KEYWORDS {
		    while {[more-text]} {
			if {[next-text-is .SH rest]} {
			    backup-text 1;
			    return;
			}
			set more [next-text];
			if {[is-a-directive $more]} {
			    manerror "$more";
			    backup-text 1;
			    return;
			}
			foreach key [split $more ,] {
			    lappend manual(keyword-[string trim $key]) [list $manual(name) $manual(wing-file)/$manual(tail).toc.html];
			}
			man-puts $more;
		    }
		    return;
		}
		DESCRIPTION {
		    if {"$manual(name)" == {options}} {
			if { ! [match-text @i1 .PP @i2 .DS @i3 .DE @i4 {.ta 4c}]} {
			    fatal "options options botch:\n[expand-next-text 8]";
			}
			man-puts $i1<P>\n$i2\n<PRE>$i3</PRE>\n$i4;
			backup-text 1;
			output-widget-options;
			return;
		    }
		}
		{STANDARD OPTIONS} {
		    if { ! [match-text .PP .nf {.ta 4c 8c 12c}]} {
			fatal "standard options botch 1:\n[next-text]\n[next-text]\n[next-text]";
		    }
		    man-puts <DL>;
		    lappend manual(section-toc) <UL>;
		    foreach opt [next-text] {
			if { ! [regexp {^<B>([a-zA-Z0-9]+)</B>$} $opt all opt]} { error "not option: $opt" }
			man-puts "<DT><B>[std-option-toc $opt]</B>";
		    }
		    man-puts </DL>;
		    lappend manual(section-toc) </UL>;
		    if { ! [match-text .fi .PP]} {
			fatal "standard options botch 2";
		    }
		    backup-text 1;
		    return;
		}
		{WIDGET-SPECIFIC OPTIONS} {
		    output-widget-options;
		    return;
		}
	    }
	    if {[next-text-is .IP rest]} {
		output-IP-list .SH .IP $rest;
		return;
	    }
	    if {[next-text-is .PP rest]} {
		return;
	    }
	    return;
	}
	.IP {
	    output-IP-list .IP .IP $rest;
	    return;
	}
	.PP {
	    man-puts <P>;
	}
	.RS {
	    output-RS-list;
	    return;
	}
	.RE {
	    manerror "unexpected .RE";
	    return;
	}
	.br {
	    man-puts <BR>;
	    return;
	}
	.DE {
	    manerror "unexpected .DE";
	    return;
	}
	.DS {
	    if {[next-text-is .ta rest]} {
		;
	    }
	    if {[match-text @stuff .DE]} {
		man-puts <PRE>$stuff</PRE>;
	    } elseif {[match-text .fi @ul1 @ul2 .nf @stuff .DE]} {
		man-puts "<PRE>[lindex $ul1 1][lindex $ul2 1]\n$stuff</PRE>";
	    } else {
		manerror "unexpected .DS format:\n[expand-next-text 2]";
	    }
	    return;
	}
	.sp {
	    man-puts <P>;
	}
	.ta {
	    # these are tab stop settings for short tables
	    switch -exact $manual(name):$manual(section) {
		{bind:MODIFIERS} -
		{bind:EVENT TYPES} -
		{bind:BINDING SCRIPTS AND SUBSTITUTIONS} -
		{expr:OPERANDS} -
		{expr:MATH FUNCTIONS} -
		{history:DESCRIPTION} -
		{history:HISTORY REVISION} -
		{switch:DESCRIPTION} -
		{upvar:DESCRIPTION} {
		    return;			# fix.me
		}
		default {
		    manerror "ignoring $line";
		}
	    }
	}
	.nf {
	    if {[match-text @more .fi]} {
		foreach more [split $more \n] {
		    man-puts $more<BR>;
		}
	    } elseif {[match-text .RS @more .RE .fi]} {
		man-puts <UL>;
		foreach more [split $more \n] {
		    man-puts $more<BR>;
		}
		man-puts </UL>;
	    } elseif {[match-text .RS @more .RS @more2 .RE .RE .fi]} {
		man-puts <UL>;
		foreach more [split $more \n] {
		    man-puts $more<BR>;
		}
		man-puts <UL>;
		foreach more2 [split $more2 \n] {
		    man-puts $more2<BR>;
		}
		man-puts </UL>;
		man-puts </UL>;
	    } elseif {[match-text .RS @more .RS @more2 .RE @more3 .RE .fi]} {
		man-puts <UL>;
		foreach more [split $more \n] {
		    man-puts $more<BR>;
		}
		man-puts <UL>;
		foreach more2 [split $more2 \n] {
		    man-puts $more2<BR>;
		}
		man-puts </UL>;
		foreach more3 [split $more3 \n] {
		    man-puts $more3<BR>;
		}
		man-puts </UL>;
	    } elseif {[match-text .sp .RS @more .RS @more2 .sp .RE .RE .fi]} {
		man-puts <P><UL>;
		foreach more [split $more \n] {
		    man-puts $more<BR>;
		}
		man-puts <UL>;
		foreach more2 [split $more2 \n] {
		    man-puts $more2<BR>;
		}
		man-puts </UL></UL><P>;
	    } elseif {[match-text .RS .sp @more .sp .RE .fi]} {
		man-puts <P><UL>;
		foreach more [split $more \n] {
		    man-puts $more<BR>;
		}
		man-puts </UL><P>;
	    } else {
		manerror "ignoring $line";
	    }
	}
	.fi {
	    manerror "ignoring $line";
	}
	.na -
	.ad -
	.UL -
	.ne {
	    manerror "ignoring $line";
	}
	default {
	    manerror "unrecognized format directive: $line";
	}
    }
}
##
## merge copyright listings
## 
proc merge-copyrights {l1 l2} {
    foreach copyright [concat $l1 $l2] {
	if {[regexp {^Copyright +\(c\) +([0-9]+) +([A-Za-z].*)$} $copyright all date who]} {
	    lappend dates($who) $date;
	    continue;
	}
	if {[regexp {^Copyright +\(c\) +([0-9]+)-([0-9]+) +([A-Za-z].*)$} $copyright all from to who]} {
	    for {set date $from} {$date <= $to} {incr date} {
		lappend dates($who) $date;
	    }
	    continue;
	}
	if {[regexp {^Copyright +\(c\) +([0-9]+), *([0-9]+) +([A-Za-z].*)$} $copyright all date1 date2 who]} {
	    lappend dates($who) $date1 $date2;
	    continue;
	}
	puts "oops: $copyright";
    }
    foreach who [array names dates] {
	set list [lsort $dates($who)];
	if {[llength $list] == 1 || [lindex $list 0] == [lrange $list end end]} {
	    lappend merge "Copyright (c) [lindex $list 0] $who";
	} else {
	    lappend merge "Copyright (c) [lindex $list 0]-[lrange $list end end] $who";
	}
    }
    return [lsort $merge];
}
    
proc makedirhier {dir} {
    if { ! [file isdirectory $dir]} {
	makedirhier [file dirname $dir];
	if { ! [file isdirectory $dir]} {
	    if {[catch {exec mkdir $dir} error]} {
		error "cannot create directory $dir: $error";
	    }
	}
    }
}
    
##
## foreach of the man directories specified by args
## convert manpages into hypertext in the directory
## specified by html.
##
proc make-man-pages {html args} {
    global env manual;
    makedirhier $html;
    if { ! [file isdirectory $html]} {
	exec mkdir $html;
    }
    set manual(short-toc-n) 1;
    set manual(short-toc-fp) [open $html/contents.html w];
    puts $manual(short-toc-fp) "<HTML><HEAD><TITLE>Tcl7.4/Tk4.0 Manual</TITLE></HEAD>"
    puts $manual(short-toc-fp) "<BODY><HR><H3>Tcl7.4/Tk4.0 Manual</H3><HR><DL>";
    set manual(merge-copyrights) {}
    foreach arg $args {
	set manual(wing-glob) [lindex $arg 0];
	set manual(wing-name) [lindex $arg 1];
	set manual(wing-file) [lindex $arg 2];
	set manual(wing-description) [lindex $arg 3];
	set manual(wing-copyrights) {};
	makedirhier $html/$manual(wing-file);
	set manual(wing-toc-fp) [open $html/$manual(wing-file)/contents.html w];
	# whistle
	puts stderr "scanning section $manual(wing-name)";
	# put the entry for this section into the short table of contents
	puts $manual(short-toc-fp) "<DT><A HREF=\"$manual(wing-file)/contents.html\">$manual(wing-name)</A><DD>$manual(wing-description)";
	# initialize the wing table of contents
	puts $manual(wing-toc-fp) "<HTML><HEAD><TITLE>$manual(wing-name) Manual</TITLE></HEAD>"
	puts $manual(wing-toc-fp) "<BODY><HR><H3>$manual(wing-name)</H3><HR>";
	# initialize the short table of contents for this section
	set manual(wing-toc) {};
	# initialize the man directory for this section
	makedirhier $html/$manual(wing-file);
	# initialize the long table of contents for this section
	set manual(long-toc-n) 1;
	# get the manual pages for this section
	set manual(pages) [lsort [glob $manual(wing-glob)]];
	if {[lsearch -glob $manual(pages) */options.n] >= 0} {
	    set n [lsearch $manual(pages) */options.n];
	    set manual(pages) "[lindex $manual(pages) $n] [lreplace $manual(pages) $n $n]";
	}
	# set manual(pages) [lrange $manual(pages) 0 5];
	foreach manual(page) $manual(pages) {
	    # whistle
	    puts stderr "scanning page $manual(page)";
	    set manual(tail) [file tail $manual(page)];
	    set manual(name) [file root $manual(tail)];
	    set manual(section) {};
	    if {[lsearch {case pack-old menubar} $manual(name)] >= 0} {
		# obsolete
		manerror "discarding $manual(name)";
		continue;
	    }
	    set manual(infp) [open "$manual(page)"];
	    set manual(text) {};
	    set manual(partial-text) {};
	    set manual(.RS) 0;
	    set manual(.DS) 0;
	    set manual(stack) {};
	    set manual(section) {};
	    set manual(section-toc) {};
	    set manual(long-toc) {};
	    set manual(section-toc-n) 1;
	    set manual(copyrights) {};
	    lappend manual(all-pages) $manual(wing-file)/$manual(tail);
	    manreport 100 "$manual(name)";
	    while {[gets $manual(infp) line] >= 0} {
		manreport 100 $line;
		if {[regexp {^[`'][/\\]} $line]} {
		    if {[regexp {Copyright \(c\).*$} $line copyright]} {
			lappend manual(copyrights) $copyright;
		    }
		    # comment
		    continue;
		}
		if {"$line" == {'}} {
		    # comment
		    continue;
		}
		if {[parse-directive $line code rest]} {
		    switch -exact $code {
			.ad -
			.na -
			.so -
			.ne -
			.AS -
			.VE -
			.VS {
			    # ignore
			    continue;
			}
		    }
		    if {"$manual(partial-text)" != {}} {
			lappend manual(text) [process-text $manual(partial-text)];
			set manual(partial-text) {};
		    }
		    switch -exact $code {
			.SH {
			    if {[llength $rest] == 0} {
				gets $manual(infp) rest;
			    }
			    lappend manual(text) ".SH [unquote $rest]";
			}
			.TH {
			    regsub {""} $rest 4.0 rest;
			    lappend manual(text) "$code [unquote $rest]";
			}
			.HS -
			.UL -
			.ta {
			    lappend manual(text) "$code [unquote $rest]";
			}
			.BS -
			.BE -
			.br -
			.fi -
			.sp -
			.nf {
			    if {"$rest" != {}} {
				manerror "unexpected argument: $line";
			    }
			    lappend manual(text) $code;
			}
			.AP {
			    lappend manual(text) [concat .IP [process-text "[lindex $rest 0] \\fB[lindex $rest 1]\\fR ([lindex $rest 2])"]];
			}
			.IP {
			    if {[llength $rest] > 1} {
				set rest [unquote [lrange $rest 0 0]];
			    }
			    lappend manual(text) ".IP [process-text $rest]";
			}
			.TP {
			    lappend manual(text) ".IP [process-text [gets $manual(infp)]]";
			}
			.PP -
			.LP {
			    lappend manual(text) {.PP};
			}
			.RS {
			    incr manual(.RS);
			    lappend manual(text) $code;
			}
			.RE {
			    incr manual(.RS) -1;
			    lappend manual(text) $code;
			}
			.DS {
			    incr manual(.DS);
			    lappend manual(text) $code;
			}
			.DE {
			    incr manual(.DS) -1;
			    lappend manual(text) $code;
			}
			.de {
			    while {[gets $manual(infp) line] >= 0} {
				if {[regexp {^\.\.} $line]} {
				    break;
				}
			    }
			}
			.. {
			    error "found .. outside of .de";
			}
			default {
			    manerror "unrecognized format directive: $line";
			}
		    }
		} else {
		    if {"$manual(partial-text)" == {}} {
			set manual(partial-text) $line;
		    } else {
			append manual(partial-text) \n$line;
		    }
		}
	    }
	    if {"$manual(partial-text)" != {}} {
		lappend manual(text) [process-text $manual(partial-text)];
	    }
	    close $manual(infp);
	    # fixups
	    if {$manual(.RS) != 0} {
		if {"$manual(name)" != {selection}} {
		    puts "unbalanced .RS .RE";
		}
	    }
	    if {$manual(.DS) != 0} {
		puts "unbalanced .DS .DE";
	    }
	    # output conversion
	    open-text;
	    if {[next-text-is .HS rest]} {
		man-puts <HTML><HEAD>
		set manual($manual(name)-title) $rest;
		man-puts "<TITLE>[lrange $rest 1 end] [lindex $rest 0] manual page</TITLE>";
		man-puts </HEAD><BODY>
		while {[more-text]} {
		    set line [next-text];
		    if {[is-a-directive $line]} {
			output-directive $line;
		    } else {
			man-puts $line;
		    }
		}
		man-puts <HR><PRE>;
		foreach copyright $manual(copyrights) {
		    man-puts "<A HREF=\"../copyright.html\">Copyright</A> &#169; [lrange $copyright 2 end]";
		}
		man-puts "<A HREF=\"../copyright.html\">Copyright</A> &#169; 1995 Roger E. Critchlow Jr.";
		man-puts </PRE></BODY></HTML>;
		set manual(wing-copyrights) [merge-copyrights $manual(wing-copyrights) $manual(copyrights)];
	    } elseif {[next-text-is .TH rest]} {
		man-puts <HTML><HEAD>
		set manual($manual(name)-title) $rest;
		man-puts "<TITLE>[lrange $rest 4 end] - [lindex $rest 0] manual page</TITLE>";
		man-puts </HEAD><BODY>
		while {[more-text]} {
		    set line [next-text];
		    if {[is-a-directive $line]} {
			output-directive $line;
		    } else {
			man-puts $line;
		    }
		}
		man-puts <HR><PRE>;
		foreach copyright $manual(copyrights) {
		    man-puts "<A HREF=\"../copyright.html\">Copyright</A> &#169; [lrange $copyright 2 end]";
		}
		man-puts "<A HREF=\"../copyright.html\">Copyright</A> &#169; 1995 Roger E. Critchlow Jr.";
		man-puts </PRE></BODY></HTML>;
		set manual(wing-copyrights) [merge-copyrights $manual(wing-copyrights) $manual(copyrights)];
	    } else {
		manerror "no .HS or .TH record found";
	    }
	    #
	    # make the long table of contents for this page
	    #
	    set manual(long-toc-fp) [open $html/$manual(wing-file)/$manual(tail).toc.html w];
	    puts $manual(long-toc-fp) "<HTML><HEAD><TITLE>$manual(tail) Contents</TITLE></HEAD>"
	    puts $manual(long-toc-fp) "<BODY>";
	    foreach line $manual(long-toc) {
		puts $manual(long-toc-fp) $line;
	    }
	    puts $manual(long-toc-fp) <UL>;
	    foreach line $manual(section-toc) {
		puts $manual(long-toc-fp) $line;
	    }
	    puts $manual(long-toc-fp) "</UL></BODY></HTML>";
	    close $manual(long-toc-fp);
	}

	#
	# make the wing table of contents for the section
	#
	set width 0;
	foreach name $manual(wing-toc) {
	    if {[string length $name] > $width} {
		set width [string length $name];
	    }
	}
	incr width 2;
	set perline [expr 80 / $width];
	puts $manual(wing-toc-fp) <PRE>;
	set n 0;
	foreach name [lsort $manual(wing-toc)] {
	    set tail $manual(name-$name);
	    if {[llength $tail] > 1} {
		manerror "$name is defined in more than one file: $tail";
		set tail [lindex $tail [expr [llength $tail]-1]];
	    }
	    set tail [file tail $tail];
	    puts -nonewline $manual(wing-toc-fp) "<A HREF=\"$tail.toc.html\">$name</A>";
	    if {[incr n] == $perline} {
		puts $manual(wing-toc-fp) {};
		set n 0;
	    } else {
		for {set i [string length $name]} {$i < $width} {incr i} {
		    puts -nonewline $manual(wing-toc-fp) { };
		}
	    }
	}

	#
	# insert wing copyrights
	#
	puts $manual(wing-toc-fp) "</PRE><HR><PRE>"
	foreach copyright $manual(wing-copyrights) {
	    puts $manual(wing-toc-fp) "<A HREF=\"../copyright.html\">Copyright</A> &#169; [lrange $copyright 2 end]";
	}
	puts $manual(wing-toc-fp) "<A HREF=\"../copyright.html\">Copyright</A> &#169; 1995 Roger E. Critchlow Jr.";
	puts $manual(wing-toc-fp) "</PRE></BODY></HTML>";
	close $manual(wing-toc-fp);
	set manual(merge-copyrights) [merge-copyrights $manual(merge-copyrights) $manual(wing-copyrights)];
    }

    ##
    ## build the keyword index.
    ## FIX.ME - needs to be split into pieces
    ##
    makedirhier $html/Keywords
    puts $manual(short-toc-fp) {<DT><A HREF="Keywords/contents.html">Keywords</A><DD>The keywords from the Tcl/Tk man pages.};
    set keyfp [open $html/Keywords/contents.html w];
    puts $keyfp "<HTML><HEAD><TITLE>Tcl/Tk Keywords</TITLE></HEAD>"
    puts $keyfp "<BODY><HR><H3>Tcl/Tk Keywords</H3><HR><UL>"
    proc strcasecmp {a b} { return [string compare [string tolower $a] [string tolower $b]]; }
    set keys [lsort -command strcasecmp [array names manual keyword-*]];
    set nkeys [llength $keys];
    set seg 15;
    for {set i $seg} {$i <= 25} {incr i} {
	if {($nkeys % $i) > ($nkeys % $seg)} {
	    set seg $i;
	}
    }
    for {set i 0} {$i < $nkeys} {incr i $seg} {
	set keyseg [lrange $keys $i [expr $i+$seg-1]];
	set seg1 [string range [lindex $keyseg 0] 8 end];
	set segn [string range [lindex $keyseg [expr [llength $keyseg]-1]] 8 end];
	puts $keyfp "<LI><A HREF=\"seg$i.html\">$seg1 <I>through</I> $segn</A>"
	set segfp [open $html/Keywords/seg$i.html w];
	puts $segfp "<HTML><HEAD><TITLE>Tcl/Tk Keywords $seg1 .. $segn</TITLE></HEAD>"
	puts $segfp "<BODY><HR><H3>Tcl/Tk Keywords $seg1 <I>through</I> $segn</H3><HR><UL>"
	foreach key $keyseg {
	    puts $segfp "<LI>[string range $key 8 end] - ";
	    set refs {};
	    foreach man $manual($key) {
		set name [lindex $man 0];
		set file [lindex $man 1]
		lappend refs "<A HREF=\"../$file\">$name</A>";
	    }
	    puts $segfp [join $refs {, }];
	}
	puts $segfp </UL><HR><PRE>;
	# insert merged copyrights
	foreach copyright $manual(merge-copyrights) {
	    puts $segfp "<A HREF=\"copyright.html\">Copyright</A> &#169; [lrange $copyright 2 end]";
	}
	puts $segfp "<A HREF=\"copyright.html\">Copyright</A> &#169; 1995 Roger E. Critchlow Jr.";
	puts $segfp </PRE><HR></BODY></HTML>
	close $segfp;
    }
    puts $keyfp </UL><HR><PRE>;
    # insert merged copyrights
    foreach copyright $manual(merge-copyrights) {
	puts $keyfp "<A HREF=\"copyright.html\">Copyright</A> &#169; [lrange $copyright 2 end]";
    }
    puts $keyfp "<A HREF=\"copyright.html\">Copyright</A> &#169; 1995 Roger E. Critchlow Jr.";
    puts $keyfp </PRE><HR></BODY></HTML>
    close $keyfp;

    ##
    ## finish off short table of contents
    ##
    puts $manual(short-toc-fp) {<DT><A HREF="http://www.elf.org">Source</A><DD>More information about these man pages.}
    puts $manual(short-toc-fp) "</DL><HR><PRE>";
    # insert merged copyrights
    foreach copyright $manual(merge-copyrights) {
	puts $manual(short-toc-fp) "<A HREF=\"copyright.html\">Copyright</A> &#169; [lrange $copyright 2 end]";
    }
    puts $manual(short-toc-fp) "<A HREF=\"copyright.html\">Copyright</A> &#169; 1995 Roger E. Critchlow Jr.";
    puts $manual(short-toc-fp) "</PRE></BODY></HTML>";
    close $manual(short-toc-fp);

    ##
    ## output man pages
    ##
    unset manual(section);
    foreach path $manual(all-pages) {
	set manual(wing-file) [file dirname $path];
	set manual(tail) [file tail $path];
	set manual(name) [file root $manual(tail)];
	puts stderr "rescanning page $manual(name)";
	if { ! [info exists manual(output-$manual(wing-file)-$manual(name))] } {
	    puts "cannot find manual(output-$manual(wing-file)-$manual(name))";
	    foreach name [array name manual output-*] {
		puts $name;
	    }
	}
	set manual(outfp) [open $html/$manual(wing-file)/$manual(tail).html w];
	foreach text $manual(output-$manual(wing-file)-$manual(name)) {
	    puts $manual(outfp) [insert-cross-references $text];
	}
	close $manual(outfp);
    }
    return {};
}

set usercmddesc {The interpreters which implement Tcl and Tk.}
set tclcmddesc {The commands which the <B>tclsh</B> interpreter implements.}
set tkcmddesc {The additional commands which the <B>wish</B> interpreter implements.}
set tcllibdesc {The C functions which a Tcl extended C program may use.}
set tklibdesc {The additional C functions which a Tk extended C program may use.}
		
if {[catch {
    make-man-pages $webdir \
	"$tcltkdir/{$tkdir,$tcldir}/doc/*.1 {Tcl/Tk Applications} UserCmd {$usercmddesc}" \
	"$tcltkdir/$tcldir/doc/*.n {Tcl Commands} TclCmd {$tclcmddesc}" \
	"$tcltkdir/$tkdir/doc/*.n {Tk Commands} TkCmd {$tkcmddesc}" \
	"$tcltkdir/$tcldir/doc/*.3 {Tcl Library} TclLib {$tcllibdesc}" \
	"$tcltkdir/$tkdir/doc/*.3 {Tk Library} TkLib {$tklibdesc}"
} error]} {
    puts $error\n$errorInfo;
}

